/*
 * LPE class : la Ligne de Partage des Eaux( Watershed)
 * @file: LPE.cpp 
 * @created: 2011.01.10 
 * @author: Youssef  El merabet
 *
 * @description: Cet algorithme reprend le principe du watershed binaire : l'utilisation d'une liste de pixels à explorer.
 *               La méthode "process" est une boucle qui fait monter le niveau de l'eau. A chaque tour de boucle,
 * 		 cette méthode fait successivement 2 tâches :
 *		 --->  étendre les bassins existant (méthodes nextPixel + extend)
 *		 --->  créer des nouveaux bassins (méthodes findSeed)
 *              A la sortie de la boucle, on construit le tableau contenant les lignes de séparation des eaux.
 * @see Watersheds in Digital Spaces: An Efficient Algorithm Based on Immersion Simulations 1991
 */

#include <pandore.h>
#include <stdio.h>
#include <list>
#include <vector>
#include <iostream>
#include <math.h>

using namespace std;
using namespace pandore;


struct Pixel {
    Pixel(const Point2d &p, const int l, bool e) : pt(p), level(l), found(e) {
    }

    Pixel(bool e) : found(e) {
    }

    Point2d pt;
    int level;
    bool found;
};


// find a seed ( = unassigned pixel ) at the specified level
Pixel findSeed(const Img2duc &ims, Reg2d &rmap, int level, int yoffset) {
    Point2d pt;
    for (pt.y = yoffset; pt.y < ims.Height(); pt.y++) {
        for (pt.x = 0; pt.x < ims.Width(); pt.x++) {
            if (ims[pt] == level && rmap[pt] == 0) {
                return Pixel(pt, level, true);
            }
        }
    }
    return NULL;
}


// return the next pixel to explore
Pixel nextPixel(vector <std::list<Pixel>> &ListPixel, int level, const int step, const int GRAYLEVEL) {
    // return the first pixel found in the ListPixel p;
    for (int i = level; i < level + step && i < GRAYLEVEL; i++) {
        if (!ListPixel[i].empty()) {
            Pixel pixel = ListPixel[i].front();
            ListPixel[i].erase(ListPixel[i].begin());
            return pixel;
        }
    }
    return NULL;
}


// explore the 8 neighbors of a pixel and set the region
void extend(const Img2duc &ims, vector <std::list<Pixel>> &ListPixel, Pixel pix, Reg2d &rmap, const int WATERSHED) {

    Point2d pt;
    int region = rmap[pix.pt];

    // this pixel is a watershed => cannot extend it
    if (region == WATERSHED) { return; }

    // for each neighbor pixel
    for (int k = 0; k < 8; k++) {
        pt = pix.pt + v8[k];

        if (!ims.Hold(pt)) { continue; }

        // level and region of neighbor
        int vk = ims[pt];
        int rk = rmap[pt];

        // neighbor is a watershed => ignore
        if (rk == -1) { continue; }

        // neighbor as no region assigned => set it
        if (rk == 0) {
            rmap[pt] = region;
            ListPixel[vk].push_back(Pixel(pt, vk, true));
            continue;
        }

        // neighbor is assigned to the same region => nothing to do
        if (rk == region) { continue; }

        // neighbor is assigned to another region => it's a watershed
        rmap[pt] = -1;
    }
}


/**
 *  step = nombre des niveaux voisins a explorer à chaque remplissage
 */
Errc pWatershedLine(const Img2duc &ims, Imc2duc &imd, int step) {
    Point2d pt;
    int level = 0, yoffset = 0, maxid = 0, WATERSHED = -1, GRAYLEVEL = 256;
    // region map
    Reg2d rmap(ims.Props());

    vector <std::list<Pixel>> ListPixel(GRAYLEVEL);


    while (level < GRAYLEVEL) {
        // extend region by exploring neighbors of known pixels
        while (true) {
            Pixel pix = nextPixel(ListPixel, level, step, GRAYLEVEL);
            if (!pix.found) { break; }
            extend(ims, ListPixel, pix, rmap, WATERSHED);
        }

        // find a new seed for this level
        Pixel seed = findSeed(ims, rmap, level, yoffset);
        if (seed.found) {
            // create and assign a new region to this seed
            rmap[seed.pt] = (++maxid);
            yoffset = seed.pt.y;

            // add this seed to the list of pixel to explore
            ListPixel[level].push_back(seed);
        } else {
            // no more seed for this level -> next level
            level++;
            yoffset = 0;
        }
    }

    // build the watershed map
    for (pt.y = 0; pt.y < ims.Height(); ++pt.y) {
        for (pt.x = 0; pt.x < ims.Width(); ++pt.x) {
            if (rmap[pt] == (Ulong) WATERSHED) {
                imd.X[pt] = 255;
            }
        }
    }

    return SUCCESS;
}


#ifdef MAIN

/*
 * Modify only the following constants, and the operator switches.
 */
#define	USAGE	"usage: %s step  [-m mask]  [im_in|-] [im_out|-]"
#define	PARC	1  // Number of parameters
#define	FINC	1  // Number of input images
#define	FOUTC	1  // Number of output images
#define	MASK	1  // Level of masking

int main( int argc, char *argv[] ) {
   Errc result;              // The result code of the execution.
   Pobject* mask;            // The region map.
   Pobject* objin[FINC+1];   // The input objects;
   Pobject* objs[FINC+1];    // The source objects masked.
   Pobject* objout[FOUTC+1]; // The output object.
   Pobject* objd[FOUTC+1];   // The result object of the execution.
   char* parv[PARC+1];       // The input parameters.

   ReadArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,parv,USAGE,MASK);
   // verifier toutes les entrees
   if ( objs[0]->Type() == Po_Img2duc) {
      
      Img2duc* const ims=(Img2duc*)objs[0];
      objd[0]=new Imc2duc(ims->Size());
      Imc2duc* const imd=(Imc2duc*)objd[0];
    
      
      result = pWatershedLine(*ims,*imd,(int)atoi(parv[0]));
      goto end;
   }
    
   
  {
     PrintErrorFormat(objin,FINC);
     result = FAILURE;
  }	

end:
  if (result)
    WriteArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,MASK);
  Exit(result);
  return 0;
}
#endif

