/*
* Statistical Region Merging (SRM) is a fast and robust algorithm to segment an image into regions
 * of similar intensity or color.
* The idea is to start with one region per pixel and then applying a statistical test on neighboring regions (in ascending order of intensity differences) whether the mean intensities are sufficiently similar enough to be merged.
*
* L’idée de base de cette méthode est de débuter avec une région par pixel, puis d’appliquer un test statistique sur les régions voisines (dans l’ordre croissant des différences d’intensité) afin de vérifier si les intensités moyennes sont suffisamment proches pour être fusionnées
* SRM a été proposée par Nock and Nielsen en 2004
* SRM intègre à la fois une technique de segmentation par croissance de régions et par fusion.
*
* Le paramètre Q représente la complexité statistique de l’image et contrôle le niveau de segmentation.
*/

#include <pandore.h>
#include <cstdio>
#include <cstdlib>
#include <stdio.h>
#include <list>
#include <vector>
#include <string.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include "statisticalmerging/segment-graph.h"

using namespace pandore;

template<typename T>
void initialisation(const Imx3d <T> &ims, Region *lut) {
    int numBands = ims.Bands();

    Ulong nb;

    for (int b = 0; b < numBands; ++b) {
        T *ps = ims.Vector(b);
        nb = 0;
        const T *pf = ims.Vector(b) + ims.VectorSize();
        for (; ps < pf; ps++) {
            if (b == 0) { lut[nb].r = (Uchar) * ps; }
            if (b == 1) { lut[nb].g = (Uchar) * ps; }
            if (b == 2) { lut[nb].b = (Uchar) * ps; }
            if (numBands == 1) {
                lut[nb].g = 0;
                lut[nb].b = 0;
            }
            lut[nb].nb = 1;
            nb++;
        }
    }
}


template<typename T>
Errc segmentation(const Imx3d <T> &ims, Reg2d &imd, double Q, int minimumRegionArea) {
    int w, h, index, npair;
    int r1, g1, b1;
    int r2, g2, b2;
    Region *lut;

    w = ims.Width();
    h = ims.Height();


    if (ims.ColorSpace() != RGB) {
        std::cerr << "Error: Bad color space, RGB expected!" << std::endl;
        return FAILURE;
    }


    lut = new Region[w * h];
    initialisation(ims, lut);


    int cpair = 0;
    // Consider C4-connectivity here
    npair = 2 * (w - 1) * (h - 1) + (h - 1) + (w - 1);
    // Construction de graph
    edge *edges = new edge[npair];

    for (int y = 0; y < h - 1; y++) {
        for (int x = 0; x < w - 1; x++) {
            index = y * w + x;

            // C4  left
            edges[cpair].r1 = index;
            edges[cpair].r2 = index + 1;

            r1 = lut[index].r;
            g1 = lut[index].g;
            b1 = lut[index].b;

            r2 = lut[index + 1].r;
            g2 = lut[index + 1].g;
            b2 = lut[index + 1].b;

            edges[cpair].diff = (int) max3(abs(r2 - r1), abs(g2 - g1), abs(b2 - b1));
            cpair++;

            // C4 below
            edges[cpair].r1 = index;
            edges[cpair].r2 = index + w;

            r2 = lut[index + w].r;
            g2 = lut[index + w].g;
            b2 = lut[index + w].b;

            edges[cpair].diff = (int) max3(abs(r2 - r1), abs(g2 - g1), abs(b2 - b1));
            cpair++;
        }
    }

    //
    // The two border lines
    //
    for (int i = 0; i < h - 1; i++) {
        index = i * w + w - 1;
        edges[cpair].r1 = index;
        edges[cpair].r2 = index + w;

        r1 = lut[index].r;
        g1 = lut[index].g;
        b1 = lut[index].b;

        r2 = lut[index + w].r;
        g2 = lut[index + w].g;
        b2 = lut[index + w].b;
        edges[cpair].diff = (int) max3(abs(r2 - r1), abs(g2 - g1), abs(b2 - b1));
        cpair++;
    }

    for (int j = 0; j < w - 1; j++) {
        index = (h - 1) * w + j;

        edges[cpair].r1 = index;
        edges[cpair].r2 = index + 1;

        r1 = lut[index].r;
        g1 = lut[index].g;
        b1 = lut[index].b;

        r2 = lut[index + 1].r;
        g2 = lut[index + 1].g;
        b2 = lut[index + 1].b;
        edges[cpair].diff = (int) max3(abs(r2 - r1), abs(g2 - g1), abs(b2 - b1));

        cpair++;
    }

    // segmentation
    universe *u = segment_graph(cpair, edges, lut, Q, minimumRegionArea, h, w);
    delete[] edges;


    int label = 1;
    int *nbe = new int[h * w];
    for (int i = 0; i < h * w; i++) {
        nbe[i] = 0;
    }

    for (int y = 0; y < h; y++) {
        for (int x = 0; x < w; x++) {
            int comp = u->find(y * w + x);
            if (nbe[comp] == 0) {
                nbe[comp] = label++;
            }

            imd[y][x] = nbe[comp];
        }
    }

    imd.Labels(u->num_sets());
    delete u;
    return SUCCESS;
}


#ifdef MAIN
#define USAGE   "usage: %s Q mra [im_in|-] [im_out|-]"
#define PARC    2
#define FINC    1
#define FOUTC   1
#define MASK    1

int main(int argc,char* argv[]) {

    Errc result;		        // The result code of the execution.
    Pobject *mask;		        // The mask.
    Pobject *objin[FINC+1];		// The input objects;
    Pobject *objs[FINC+1];		// The source objects masked by mask.
    Pobject *objout[FOUTC+1];	// The output object unmasked by mask.
    Pobject *objd[FOUTC+1];		// The result object.
    char *parv[PARC+1];			// The input parameters.

    ReadArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,parv,USAGE,MASK);


       if (objs[0]->Type() == Po_Imc2duc ) {

          Imc2duc* const ims=(Imc2duc*)objs[0];
          objd[0]=new Reg2d(ims->Props());
          Reg2d* const imd=(Reg2d*)objd[0];

         result = segmentation(*ims,*imd,(double)atoi(parv[0]), atoi(parv[1]));

         goto end;
      }

       if (objs[0]->Type() == Po_Img2duc ) {

          Img2duc* const ims=(Img2duc*)objs[0];
          objd[0]=new Reg2d(ims->Props());
          Reg2d* const imd=(Reg2d*)objd[0];

         result = segmentation(*ims,*imd,(double)atoi(parv[0]), atoi(parv[1]));

         goto end;
      }


  {
     PrintErrorFormat(objin,FINC);
     result = FAILURE;
  }

end:
  if (result)
    WriteArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,MASK);
  Exit(result);
  return 0;
}

#endif
