
#ifndef SEGMENT_GRAPH
#define SEGMENT_GRAPH

#include <algorithm>
#include <cmath>
#include "disjoint-set.h"

using namespace pandore;

#define Smallregion(w, h) (0.0001*w*h)

#define Logdelta(w, h) (2.0*log(6.0*w*h))

#define    gg    256.0   // number of levels in a color channel

typedef struct {
    int diff;
    int r1, r2;
} edge;

bool operator<(const edge &a, const edge &b) {
    return a.diff < b.diff;
}

struct Region {
    double r, g, b;
    int nb;
};

double min(double a, double b) {
    if (a < b) { return a; } else { return b; }
}

double max(double a, double b) {
    if (a > b) { return a; } else { return b; }
}

double max3(double a, double b, double c) {
    return max(a, max(b, c));
}

// Merge two regions
void MergeRegions(universe *u, Region *lut, int r1, int r2) {
    int nreg;
    double ravg, gavg, bavg;


    u->join(r1, r2);

    nreg = lut[r1].nb + lut[r2].nb;
    ravg = (lut[r1].nb * lut[r1].r + lut[r2].nb * lut[r2].r) / nreg;
    gavg = (lut[r1].nb * lut[r1].g + lut[r2].nb * lut[r2].g) / nreg;
    bavg = (lut[r1].nb * lut[r1].b + lut[r2].nb * lut[r2].b) / nreg;


    r1 = u->find(r1);
    lut[r1].nb = nreg;
    lut[r1].r = ravg;
    lut[r1].g = gavg;
    lut[r1].b = bavg;
}

// Merge small regions
void MergeSmallRegion(universe *u, Region *lut, edge *edges, int npair, int minimumRegionArea) {
    for (int i = 0; i < npair; i++) {
        edge *pedge = &edges[i];

        // components conected by this edge
        int r1 = u->find(pedge->r1);
        int r2 = u->find(pedge->r2);
        if (r2 != r1) {
            if ((lut[r1].nb < minimumRegionArea) || (lut[r2].nb < minimumRegionArea)) {
                MergeRegions(u, lut, r1, r2);
            }
        }
    }
}

bool MergePredicate(Region *lut, int r1, int r2, double Q, int h, int w) {
    double dR, dG, dB;
    double logr1, logr2;
    double dev1, dev2, dev;

    dR = (lut[r1].r - lut[r2].r);
    dR *= dR;

    dG = (lut[r1].g - lut[r2].g);
    dG *= dG;

    dB = (lut[r1].b - lut[r2].b);
    dB *= dB;

    logr1 = min(gg, lut[r1].nb) * log(1.0 + lut[r1].nb);
    logr2 = min(gg, lut[r2].nb) * log(1.0 + lut[r2].nb);

    dev1 = ((gg * gg) / (2.0 * Q * lut[r1].nb)) * (logr1 + (double) Logdelta(w, h));
    dev2 = ((gg * gg) / (2.0 * Q * lut[r2].nb)) * (logr2 + (double) Logdelta(w, h));

    dev = dev1 + dev2;

    return ((dR < dev) && (dG < dev) && (dB < dev));
}


edge *BucketSort(edge *a, int n) {
    int i;
    int *nbe;
    int *cnbe;
    edge *b = new edge[n];

    nbe = new int[256];
    cnbe = new int[256];

    for (i = 0; i < 256; i++) {
        nbe[i] = 0;
    }
    // class all elements according to their family
    for (i = 0; i < n; i++) { nbe[a[i].diff]++; }
    // cumulative histogram
    cnbe[0] = 0;
    for (i = 1; i < 256; i++) {
        cnbe[i] = cnbe[i - 1] + nbe[i - 1];
    } // index of first element of category i

    // allocation
    for (i = 0; i < n; i++) {
        b[cnbe[a[i].diff]++] = a[i];
    }

    return b;
}


/*
 * Segment a graph
 *
 * Returns a disjoint-set forest representing the segmentation.
 *
 * num_vertices: number of vertices in graph.
 * npair: number of edges in graph
 * edges: array of edges.
 */
universe *segment_graph(int npair, edge *edges, Region *lut, double Q, int minimumRegionArea, int h, int w) {
    int num_vertices = h * w;
    // Sort the edges according to the maximum color channel difference
    //
    edges = BucketSort(edges, npair);

    // make a disjoint-set forest
    universe *u = new universe(num_vertices);


    // for each edge, in non-decreasing weight order...
    for (int i = 0; i < npair; i++) {
        edge *pedge = &edges[i];

        // components conected by this edge
        int r1 = u->find(pedge->r1);
        int r2 = u->find(pedge->r2);
        if (r1 != r2) {
            if ((r1 != r2) && (MergePredicate(lut, r1, r2, Q, h, w))) {
                MergeRegions(u, lut, r1, r2);
            }
        }
    }

    MergeSmallRegion(u, lut, edges, npair, minimumRegionArea);

    return u;
}

#endif
